from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from functools import lru_cache, cache
from typing import Iterable

import matplotlib.pyplot as plt
import scipy.interpolate as si
import scipy.optimize as so
import scipy.spatial as ss
import scipy.special as se

import numpy as np

from .interpolate import TemperatureInterpolator, SGTE, PolyFit, RedlichKister

from scipy.constants import Boltzmann, eV

kB = Boltzmann / eV


def S(c):
    return kB * (se.entr(c) + se.entr(1 - c))


def Sprime(c):
    with np.errstate(divide="ignore"):
        s = -kB * (np.log(c / (1 - c)))
    s[np.isclose(c, 0)] = +np.inf
    s[np.isclose(c, 1)] = -np.inf
    return s


def c_from_dmu(dmu, T, e_defect):
    return 1 / (1 + np.exp(-(dmu - e_defect) / kB / T))


@dataclass(frozen=True)
class Phase(ABC):
    """
    Represents a phase in a binary phase diagram.
    """

    name: str

    @abstractmethod
    def semigrand_potential(self, T, dmu):
        """
        Calculate the semigrand potential of the phase.
        """
        pass

    @abstractmethod
    def concentration(self, T, dmu):
        """
        Concentration of the phase at the given state.
        """
        pass

    def __repr__(self):
        return f'{type(self).__name__}("{self.name}")'

    __str__ = __repr__


@dataclass(frozen=True)
class AbstractLinePhase(Phase):
    """Base class for fixed concentration phases.

    Required overloads are :meth:`.AbstractLinePhase.line_concentration` and
    :meth:`.AbstractLinePhase.line_free_energy`.
    """

    @property
    @abstractmethod
    def line_concentration(self):
        pass

    @abstractmethod
    def line_free_energy(self, T):
        pass

    def free_energy(self, T, c):
        return self.line_free_energy(T)

    def concentration(self, T, dmu):
        ones = np.ones(np.broadcast(T, dmu).shape)
        if ones.shape != ():
            return self.line_concentration * ones
        else:
            return self.line_concentration

    def semigrand_potential(self, T, dmu):
        f = self.line_free_energy(T)
        return f - self.line_concentration * dmu


@dataclass(frozen=True)
class LinePhase(AbstractLinePhase):
    """
    Simple phase with a fixed concentration and temperature independent entropy.
    """

    fixed_concentration: float
    line_energy: float
    line_entropy: float = 0

    @property
    def line_concentration(self):
        return self.fixed_concentration

    def line_free_energy(self, T):
        if isinstance(T, np.ndarray):
            U = self.line_energy * np.ones_like(T)
        else:
            U = self.line_energy
        return U - T * self.line_entropy


@dataclass(frozen=True)
class TemperatureDependentLinePhase(AbstractLinePhase):
    """ "
    Simple phase with a fixed concentration and temperature dependent free
    energy.
    """

    fixed_concentration: float
    """The fixed concentration of the phase"""
    temperatures: Iterable[float]
    """Temperatures at which the free energy of the phase has been sampled."""
    free_energies: Iterable[float]
    """Sampled free energy of the phase has been computed."""
    interpolator: TemperatureInterpolator = SGTE(3)
    """How to interpolate to arbitrary temperatures from the samples."""
    _hash: int = field(default=0, init=False)

    def __post_init__(self, *args, **kwargs):
        def to_ro_numpy(iterable):
            a = np.array(iterable)
            a.flags.writeable = False
            return a

        object.__setattr__(self, "temperatures", to_ro_numpy(self.temperatures))
        object.__setattr__(self, "free_energies", to_ro_numpy(self.free_energies))
        # precompute hash: hashing arrays every cache lookup is too expensive
        # and we any way advertise as frozen
        object.__setattr__(
            self,
            "_hash",
            hash(
                (
                    hash(self.fixed_concentration),
                    hash(self.temperatures.tobytes()),
                    hash(self.free_energies.tobytes()),
                    hash(self.interpolator),
                )
            ),
        )

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return all(
            (
                self.fixed_concentration == other.fixed_concentration,
                np.array_equal(self.temperatures, other.temperatures),
                np.array_equal(self.free_energies, other.free_energies),
            )
        )

    @property
    @cache
    def _interpolation(self):
        return self.interpolator.fit(self.temperatures, self.free_energies)

    @property
    def line_concentration(self):
        return self.fixed_concentration

    def line_free_energy(self, T):
        return self._interpolation(T)

    def check_interpolation(self, Tl=0.9, Tu=1.1, samples=50):
        Ts = np.linspace(np.min(self.temperatures) * Tl, np.max(self.temperatures) * Tu, samples)
        (l,) = plt.plot(Ts, self.line_free_energy(Ts), label=self.name)
        # try to plot about 100 points
        n = max(int(len(self.temperatures) // 100), 1)
        plt.scatter(self.temperatures[::n], self.free_energies[::n], c=l.get_color())


def TemperatureDepandantLinePhase(*args, **kwargs):
    print("TYPO ALERT!")
    return TemperatureDependentLinePhase(*args, **kwargs)


@dataclass(frozen=True, eq=True)
class IdealSolution(Phase):
    phase1: AbstractLinePhase
    phase2: AbstractLinePhase

    def __post_init__(self, *args, **kwargs):
        phase1, phase2 = sorted((self.phase1, self.phase2), key=lambda p: p.line_concentration)
        assert phase1.line_concentration == 0 and phase2.line_concentration == 1, "Must give terminal phases!"
        # bypass frozen=True for the sake of init only
        object.__setattr__(self, "phase1", phase1)
        object.__setattr__(self, "phase2", phase2)

    def semigrand_potential(self, T, dmu):
        T = np.asarray(T)
        dmu = np.asarray(dmu)
        p1 = self.phase1
        p2 = self.phase2
        f1 = p1.line_free_energy(T)
        f2 = p2.line_free_energy(T)
        df = f2 - f1
        with np.errstate(divide='ignore', over="ignore", invalid='ignore'):
            expo = -(df - dmu) / kB / T
            phi = f1 - kB * T * np.log(1 + np.exp(expo))
            I = ~np.isfinite(phi)
            if I.any():
                if phi.shape == ():
                    phi = f2 - dmu[I]
                else:
                    phi[I] = f2 - dmu[I]
        if phi.shape == ():
            phi = phi.item()
        return phi

    def concentration(self, T, dmu):
        p1 = self.phase1
        p2 = self.phase2
        f1 = p1.line_free_energy(T)
        f2 = p2.line_free_energy(T)
        df = f2 - f1
        with np.errstate(divide='ignore', over='ignore'):
            return 1 / (1 + np.exp(+(df - dmu) / kB / T))


@dataclass(frozen=True, eq=True)
class RegularSolution(Phase):
    """
    A regular solution model phase that interpolates through a given set of line phases using Redlich-Kister
    polynomials.
    """

    phases: Iterable[AbstractLinePhase]
    """Line phases to interpolate, *must* include the terminals."""
    num_coeffs: int = 4
    """Number of Redlich-Kister coefficients for the mixing "enthalpy"; restricted to number of phases - 2."""
    add_entropy: bool = False
    """If False, assume that the free energies of the line phases already include configurational mixing entropy. If
    True add ideal mixing entropy."""

    def __post_init__(self, *args, **kwargs):
        # bypass frozen=True for the sake of init only
        object.__setattr__(self, "phases", tuple(self.phases))
        object.__setattr__(self, "num_coeffs", min(len(self.phases) - 2, self.num_coeffs))
        concs = tuple(p.line_concentration for p in self.phases)
        assert 0 in concs and 1 in concs, "Must give the terminal phases!"
        left_terminals = sum(c == 0 for c in concs)
        right_terminals = sum(c == 1 for c in concs)
        assert left_terminals == 1 and right_terminals == 1, (
            "Cannot pass multiple terminal phases of the same concentration!"
        )

    @lru_cache(maxsize=250)
    def _get_interpolation(self, T):
        cc = np.array([l.line_concentration for l in self.phases])
        ff = np.array([l.line_free_energy(T) for l in self.phases])

        # TODO: needs better naming: If the free energies of the phase objects
        # already contain the entropy of mixing, remove it here first, before
        # we try to fit the redlich kister coeffs
        if not self.add_entropy:
            ff += T * S(cc)
        return RedlichKister(self.num_coeffs).fit(cc, ff)

    def free_energy(self, T, c):
        return self._get_interpolation(T)(c) - T * S(c)

    def excess_free_energy(self, T, c):
        cc = np.linspace(0, 1)
        ff = self.free_energy(T, cc)
        f0 = ff[0]
        f1 = ff[-1]
        return si.interp1d(cc, ff - (f0 * (1 - cc) + f1 * cc), kind="cubic")(c)

    def semigrand_potential(self, T, dmu, plot=False, raw=False):
        def get_mu_c(c):
            f = self.free_energy(T, c)

            f0 = f[0]
            f1 = f[-1]
            I = f <= c * f1 + (1 - c) * f0
            fI = f[I]
            cI = c[I]

            # system is fully demixing
            if I.sum() == 2:
                M = f1 - f0
                f12 = (f0 + f1) / 2
                return (np.array([0, 0.5, 1]), np.array([f0, f12, f1]), np.array([M - 1e-3, M, M + 1e-3]))

            hull = ss.ConvexHull(list(zip(cI, fI)))
            cH, fH = hull.points[hull.vertices].T
            Is = np.argsort(cH)
            cH = cH[Is]
            fH = fH[Is]

            M = np.gradient(fH, cH)
            return cH, fH, M

        n = 50
        c, f, M = get_mu_c(np.linspace(0, 1, n))
        limit = 5e-3
        while np.median(abs(np.diff(M))) > limit and n < 5e4:
            n *= 2
            Ms = np.linspace(M.min(), M.max(), n)
            c = si.interp1d(M, c)(Ms)
            c, f, M = get_mu_c(c)
            if plot:
                plt.subplot(121)
                plt.plot(c[:-1], np.diff(M), "v", label=n)
                plt.subplot(122)
                plt.plot(c, M, ".")
        if plot and n > 50:
            plt.subplot(121)
            plt.title("Spacing of Chemical Potential Sampling")
            plt.xlabel("c")
            plt.ylabel("np.diff(mu)")
            plt.legend(title="grid points")
            plt.subplot(122)
            plt.xlabel("c")
            plt.ylabel(r"$\Delta \mu$")
            plt.show()

        p = f - c * M
        if raw:
            return f, c, M, p

        assert np.median(abs(np.diff(M))) <= limit, "Weird"

        # schon etwas dreist, aber naja
        pi = si.interp1d(
            M,
            p,
            fill_value=np.nan,
            bounds_error=False,
            # needs to be at least quadratic, otherwise we'll see
            # jumps in the numerically calculated concentration
            kind="quadratic",
        )(dmu)
        pl = self.free_energy(T, 1) - dmu * 1
        f0 = self.free_energy(T, 0)
        if not isinstance(dmu, np.ndarray):
            if np.isnan(pi):
                pi = np.inf
            x = min(pi, pl, f0)
            if isinstance(x, np.ndarray):
                x = x.item()
            return x
        pl[pl > f0] = f0
        I = np.isnan(pi)
        pi[I] = pl[I]
        if plot:
            plt.plot(M, p, "o-", label="calculated")
            plt.plot(dmu, pi, label="extrapolated")
            plt.legend()
        return pi

    def concentration(self, T, dmu):
        if not isinstance(dmu, np.ndarray):
            dmus = np.linspace(-1, 1, 5) * 1e-3 + dmu
            return self.concentration(T, dmus)[2]
        return np.clip(-np.gradient(self.semigrand_potential(T, dmu), dmu, edge_order=2), 0, 1)

    def check_interpolation(self, T=1000, samples=50):
        x = np.linspace(0, 1, samples)
        plt.plot(x, self.free_energy(T, x), label=self.name)
        for p in self.phases:
            plt.scatter(p.line_concentration, p.line_free_energy(T))


from numbers import Real


@dataclass(frozen=True, eq=True)
class InterpolatingPhase(Phase):
    """A Version of RegularSolutionPhase that does not depend on terminals.  FIXME: These two classes should be unified."""

    phases: Iterable[AbstractLinePhase]
    num_coeffs: int = None
    add_entropy: bool = False
    num_samples: int = 100
    maximum_extrapolation: float = 0

    def __post_init__(self, *args, **kwargs):
        object.__setattr__(self, "phases", tuple(self.phases))
        object.__setattr__(self, "num_coeffs", min(len(self.phases), self.num_coeffs or np.inf))

    @lru_cache(maxsize=250)
    def _get_interpolation(self, T):
        if not isinstance(T, Real):
            raise TypeError(T)
        cc = np.array([l.line_concentration for l in self.phases])
        ff = np.array([l.line_free_energy(T) for l in self.phases])

        # TODO: needs better naming: If the free energies of the phase objects
        # already contain the entropy of mixing, remove it here first, before
        # we try to fit the redlich kister coeffs
        if not self.add_entropy:
            ff += T * S(cc)
        if cc[0] == 0 and cc[-1] == 1:
            return RedlichKister(max(1, self.num_coeffs - 2)).fit(cc, ff)
        else:
            return PolyFit(self.num_coeffs).fit(cc, ff)

    def free_energy(self, T, c):
        return np.vectorize(
            lambda T, c: self._get_interpolation(T)(c) - T * S(c),
            otypes=[float]
        )(T, c)
        # return self._get_interpolation(T)(c) - T * S(c)

    def _find_phi_c(self, T, dmu):
        """Calculate potential and concentration together.

        Formally we need to solve

        phi = min_c { f(c) - c * dmu }

        but this is too slow to solve with normal optimizers from scipy and can
        get stuck in local minima.  Instead do the brute force minimization on
        a grid (self.num_samples), then refine the gridded concentrations with
        a single step of a newton-raphson like optimization.  This makes sure
        that the output concentrations are smooth and non-degenerate.
        """
        output_shape = np.broadcast_shapes(np.shape(T), np.shape(dmu))
        T = np.atleast_1d(T)[..., np.newaxis]
        dmu = np.atleast_1d(dmu)[..., np.newaxis]

        cs = [p.line_concentration for p in self.phases]
        conc = np.linspace(
            max(0, min(cs) - self.maximum_extrapolation),
            min(1, max(cs) + self.maximum_extrapolation),
            self.num_samples
        )
        ff = self.free_energy(T, conc)
        phi = ff - conc * dmu
        I = phi.argmin(axis=-1, keepdims=True)
        phi = np.take_along_axis(phi, I, axis=-1)[..., 0]
        c = conc[I[..., 0]]

        df = np.take_along_axis(
                np.gradient(ff, conc, axis=-1, edge_order=2),
                I, axis=-1
        )
        d2f = np.take_along_axis(
                np.gradient(
                    np.gradient(ff, conc, axis=-1, edge_order=2),
                    conc, axis=-1, edge_order=2
                ),
                I, axis=-1
        )
        dc = (dmu - df) / d2f
        nc = np.clip(c + dc[..., 0], 0, 1)
        phi -= (nc-c)*(dmu-df)[..., 0]
        c = nc

        c = c.reshape(output_shape)
        phi = phi.reshape(output_shape)

        if c.ndim == 0:
            c = c.item()
        if phi.ndim == 0:
            phi = phi.item()
        return phi, c

    def semigrand_potential(self, T, dmu):
        return self._find_phi_c(T, dmu)[0]

    def concentration(self, T, dmu):
        return self._find_phi_c(T, dmu)[1]

    def check_interpolation(self, T=1000, samples=50):
        x = np.linspace(0, 1, samples)
        plt.plot(x, self.free_energy(T, x))
        for p in self.phases:
            plt.scatter(p.line_concentration, p.line_free_energy(T), label=self.name)


class AbstractPointDefect(ABC):
    @abstractmethod
    def excess_free_energy(self, T):
        pass

    # @property
    # @abstractmethod
    # def excess_solutes(self):
    #     pass


@dataclass(frozen=True)
class ConstantPointDefect(AbstractPointDefect):
    """
    A point defect that adds a contribution to the free energy of a host
    lattice.

    Excess energy and entropy are assumed to be
    """

    name: str
    excess_energy: float
    # [E]_N
    excess_entropy: float
    # [S]_N
    excess_solutes: float
    # # [n]_N / N = c_defect - c_reference

    # def __init__(self, name,
    #              excess_energy, excess_solutes,
    #              sublattice, sublattice_fraction,
    #              excess_entropy=0,
    #              # unused
    #              excess_concentration=None,
    # ):
    #     super().__init__()
    #     self.storage.name = name
    #     self.storage.excess_energy = excess_energy
    #     self.storage.excess_entropy = excess_entropy
    #     # self.storage.excess_concentration = excess_concentration
    #     # [n]_N
    #     self.storage.excess_solutes = excess_solutes
    #     # sublattice index; on which sublattice in the host lattice this defect lives
    #     # just to avoid accidental overlap when combining multiple point defects
    #     self.storage.sublattice = sublattice
    #     # the fraction of sites this sub lattice makes up of the host lattice
    #     self.storage.sublattice_fraction = sublattice_fraction

    def excess_free_energy(self, T):
        return self.excess_energy - T * self.excess_entropy

    # For diagnostics only; Sublattice classes only uses excess_free_energy method

    def semigrand_potential_contribution(self, T, dmu):
        fe = self.excess_free_energy(T)
        ne = self.excess_solutes
        return -kB * T * np.log(1 + np.exp(-(fe - ne * dmu) / kB / T))

    def concentration_contribution(self, T, dmu):
        ne = self.excess_solutes
        return ne * self.defect_concentration(T, dmu)

    def defect_concentration(self, T, dmu):
        # analytical derivative of the semigrand potential above
        # c = -dphi/dmu
        # c = [n]_N eta x
        # we want to return x
        fe = self.storage.excess_energy
        ne = self.storage.excess_solutes
        return 1 / (1 + np.exp(+(fe - ne * dmu) / kB / T))


@dataclass(frozen=True)
class PointDefectSublattice:
    """
    Groups together PointDefect that live on the same sublattice within a host
    structure.
    """

    name: str
    sublattice: int
    sublattice_fraction: float
    defects: list[AbstractPointDefect]

    def _get_zes(self, T, dmu):
        fes = [d.excess_free_energy(T) for d in self.defects]
        nes = [d.excess_solutes for d in self.defects]
        return np.array([np.exp(-(fe - ne * dmu) / kB / T) for fe, ne in zip(fes, nes)])

    def semigrand_potential_contribution(self, T, dmu):
        zes = self._get_zes(T, dmu).sum(axis=0)
        dphi = -kB * T * np.log(1 + zes)
        return self.sublattice_fraction * dphi

    def concentration_contribution(self, T, dmu):
        zes = self._get_zes(T, dmu)
        nes = np.array([p.excess_solutes for p in self.defects])
        eta = self.sublattice_fraction
        return eta * sum(ne * ze for ne, ze in zip(nes, zes)) / (1 + zes.sum(axis=0))


@dataclass(frozen=True)
class PointDefectedPhase(Phase):
    """
    Phase that combines any host phase and any number of point defects in it.
    """

    line_phase: AbstractLinePhase
    """Underlying phase object of the host lattice."""
    sublattices: list[PointDefectSublattice]
    """Sublattices and their point defects."""

    def __post_init__(self, *args, **kwargs):
        # TODO check unique sublattice indices on sublattice objects (or maybe not)
        pass

    def semigrand_potential(self, T, dmu):
        phi = self.line_phase.semigrand_potential(T, dmu)
        for l in self.sublattices:
            phi += l.semigrand_potential_contribution(T, dmu)
        return phi

    def concentration(self, T, dmu):
        c = self.line_phase.line_concentration
        for d in self.sublattices:
            c += d.concentration_contribution(T, dmu)
        return c



from typing import Literal

from matplotlib.patches import Polygon
import shapely
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
from sklearn.decomposition import PCA
import numpy as np

from .calculate import get_transitions


def make_concave_poly(dd, alpha=0.1, min_c_width=1e-3, variables=["c", "T"]):
    # concave hull algo seems more stable when both variables are of the same order

    pp = dd.sort_values(variables[0])[variables].to_numpy()
    pp = pp[np.isfinite(pp).all(axis=-1)]

    refnorm = {}
    for i, var in enumerate(variables):
        refnorm[var] = pp[:, i].min(), (np.ptp(pp[:, i]) or 1)
        pp[:, i] -= refnorm[var][0]
        pp[:, i] /= refnorm[var][1]
    shape = shapely.concave_hull(shapely.MultiPoint(pp), ratio=alpha)
    # check for c-degenerate line phase
    if variables[0] == "c" and isinstance(shape, shapely.LineString):
        coords = np.asarray(shape.coords)
        if np.allclose(coords[:, 0], coords[0, 0]):
            match coords[0, 0]:
                case 0.0:
                    bias = +min_c_width / 2
                case 1.0:
                    bias = -min_c_width / 2
                case _:
                    bias = 0
            # artificially widen the line phase in c, so that we can make a
            # "normal" polygon for it.
            coords = np.concatenate(
                [
                    # inverting the order for the second half of the array, makes
                    # it so that the points are in the correct order for the
                    # polygon
                    coords[::+1] - [min_c_width / 2, 0],
                    coords[::-1] + [min_c_width / 2, 0],
                ],
                axis=0,
            )
            coords[:, 0] += bias
    else:
        coords = np.asarray(shape.exterior.coords)
    for i, var in enumerate(variables):
        coords[:, i] *= refnorm[var][1]
        coords[:, i] += refnorm[var][0]
    return Polygon(coords)


def sort_segments(df, x_col="c", y_col="T", segment_label="border_segment"):
    """
    Sorts the points in df such that they can be used as the bounding polygon of a phase in a binary diagram.

    Assumptions:
    1. df contains only data on a single, coherent phase, i.e. the c/T points are "connected"

    Algorithm:
    1. Subset the data according to the column given by `segment_label`.  These should label connected points on a single two-phase boundary. Such a subset is called a segment.
    2. Sort points in each segment by a 1D PCA. (Sorting by c or T alone fails when the segment is either vertical or horizontal.)
    3. Sort the segments so that they "easily" fit together:
        a. Pick the segment with minimum `x` as the "head"
        b. Go over all other segments, s, and:
            b0. Get the distance from endpoint of "head" to either the starting point or the end point of s
            b1. if the distance to the end point is shorter than to the starting point, invert order of s
            b2. return the minimum of both distances
        c. the segment with smallest distance to the current "head" is the next "head" and removed from the pool of segments
        d. break if no segments left
    4. return the segments in the order they were picked as "head"s.

    a) is a heuristic for "normal" phase diagrams, starting from the left (or right) we can often make a full circle.
    Picking a random segments breaks for phases that are stable at the lower or upper edge of the diagram, where we technically do not compute
    a "segment".  A "proper" fix would be to modify b to allow joining also to the start of "head" rather than just the end.
    """

    com = df[[x_col, y_col]].mean()
    norm = np.ptp(df[[x_col, y_col]], axis=0).values


    # Step 1: PCA Projection
    def pca_projection(group):
        # avoid warnings when clustering only found one or two points
        if len(group) < 2:
            return group
        pca = PCA(n_components=1)
        projected = pca.fit_transform(group[[x_col, y_col]])
        group["projected"] = projected
        return group.sort_values("projected").copy().drop("projected", axis="columns").reset_index(drop=True)

    segments = []
    for label, dd in df.groupby(segment_label):
        segments.append(pca_projection(dd))

    # initial sorting by center of mass angle
    segments = sorted(
            segments,
            key=lambda s: np.arctan2( (s[y_col].mean() - com[y_col]) / norm[1],
                                      (s[x_col].mean() - com[x_col]) / norm[0])
    )

    def start(s):
        return s.iloc[0][[x_col, y_col]]

    def end(s):
        return s.iloc[-1][[x_col, y_col]]

    def dist(p1, p2):
        return np.linalg.norm((p2 - p1) / norm)

    def flip(s):
        s.reset_index(drop=True, inplace=True)
        s.loc[:] = s.loc[::-1].reset_index(drop=True)
        return s

    head, *remaining = sorted(segments, key=lambda s: s[x_col].min())

    def find_distance(head, segment):
        head2tail = dist(end(head), start(segment))
        tail2tail = dist(end(head), end(segment))
        if tail2tail < head2tail:
            flip(segment)
            return tail2tail
        else:
            return head2tail

    segments = [head]
    while len(remaining) > 0:
        head, *remaining = sorted(remaining, key=lambda s: find_distance(head, s))
        segments.append(head)

    return pd.concat(segments, ignore_index=True)


def make_poly(td, min_c_width=1e-3, variables=["c", "T"]):
    """
    Requires a grouped dataframe from find_transitions (by phase).
    """
    if "c" in variables and np.ptp(td.c) < min_c_width:
        meanc = td.c.mean()
        Tmin = td["T"].min()
        Tmax = td["T"].max()
        return Polygon(
            [
                [meanc - min_c_width / 2, Tmin],
                [meanc + min_c_width / 2, Tmin],
                [meanc + min_c_width / 2, Tmax],
                [meanc - min_c_width / 2, Tmax],
            ]
        )
    td = td.loc[ np.isfinite(td[variables[0]]) & np.isfinite(td[variables[1]]) ]
    sd = sort_segments(td, x_col=variables[0], y_col=variables[1])
    # sd = sd.loc[ np.isfinite(sd[variables[0]]) & np.isfinite(sd[variables[1]]) ]
    return Polygon(np.transpose([sd[v] for v in variables]))


def plot_phase_diagram(
    df, alpha=0.1, element=None, min_c_width=5e-3, color_override: dict[str, str] = {}, tielines=False,
    poly_method: Literal["concave", "segments"] = 'concave',
):
    df = df.query("stable")

    # the default map
    color_map = dict(zip(df.phase.unique(), sns.palettes.SEABORN_PALETTES["pastel"]))
    # disregard overriden phases that are not present
    color_override = {p: c for p, c in color_override.items() if p in color_map}
    # if the override uses the same colors as the default map, multiple phases
    # would be mapped to the same color; so instead let's update the color map of phases that would
    # use the same color as a phase in the override to use the default colors of the overriden phases
    # instead
    duplicates_map = {c: color_map[o] for o, c in color_override.items()}
    diff = {k: duplicates_map[c] for k, c in color_map.items() if c in duplicates_map}
    color_map.update(diff | color_override)

    if "refined" in df.columns and poly_method == "segments":
        tdf = get_transitions(df)
        polys = tdf.groupby("phase").apply(
            make_poly,
            min_c_width=min_c_width,
        )
    else:
        polys = df.groupby("phase").apply(
            make_concave_poly,
            alpha=alpha,
            min_c_width=min_c_width,
        )

    ax = plt.gca()
    for i, (phase, p) in enumerate(polys.items()):
        p.zorder = 1/p.get_extents().size.prod()
        p.set_color(color_map[phase])
        p.set_edgecolor("k")
        # p.set_alpha(.8)
        p.set_label(polys.index[i])
        ax.add_patch(p)

    if tielines:
        # TODO: quite buggy and not nice; can benefit a lot from
        # get_transitions
        if "refined" in df.columns:
            tdf = get_transitions(df)
            def plot_tie(dd):
                Tmin = dd["T"].min()
                Tmax = dd["T"].max()
                di = dd.query("T==@Tmin")
                da = dd.query("T==@Tmax")
                # "artificial" segment at the border of diagram
                # we just want to plot triple lines? so #phases==3
                if len(dd.phase.unique()) in [1, 2]:
                    return
                plt.hlines(Tmin, di.c.min(), di.c.max(), color="k", zorder=-2, alpha=0.5, lw=4)
                # current marvin to past marvin: Why is that even necessary?
                if Tmin != Tmax:
                    plt.hlines(Tmax, da.c.min(), da.c.max(), color="k", zorder=-2, alpha=0.5, lw=4)

            # FIXME: WARNING reuses local var define in if branch
            tdf.groupby("border_segment").apply(plot_tie)
        else:
            # count the numbers of distinct phases per T, it changes there *must* be a triple
            # point, draw tie lines only there
            # TODO: figure out how to only draw them between the involved phases not over the whole conc range
            # the refined data points mess this up, because the phases are no longer on
            # the same grid
            chg = df.groupby("T").size().diff()
            T_tie = chg.loc[chg != 0].index[1:]  # skip first temp

            def plot_tie(dd):
                if dd["T"].iloc[0].round(3) not in T_tie.round(3):
                    return
                if len(dd) != 2:
                    return
                cl, cr = sorted(dd.c)
                plt.plot([cl, cr], dd["T"], color="k", zorder=-2, alpha=0.5, lw=4)

            df.groupby(["T", "mu"]).apply(plot_tie)

    plt.xlim(0, 1)
    plt.ylim(df["T"].min(), df["T"].max())
    plt.legend(ncols=2)
    if element is not None:
        plt.xlabel(rf"$c_\mathrm{{{element}}}$")
    else:
        plt.xlabel("$c$")
    plt.ylabel("$T$ [K]")

def get_phase_colors(phase_names, override: dict[str, str] | None = None):
    # the default map
    color_map = dict(zip(phase_names, sns.palettes.SEABORN_PALETTES["pastel"]))
    # disregard overriden phases that are not present
    override = {p: c for p, c in override.items() if p in color_map}
    # if the override uses the same colors as the default map, multiple phases
    # would be mapped to the same color; so instead let's update the color map of phases that would
    # use the same color as a phase in the override to use the default colors of the overriden phases
    # instead
    duplicates_map = {c: color_map[o] for o, c in override.items()}
    diff = {k: duplicates_map[c] for k, c in color_map.items() if c in duplicates_map}
    color_map.update(diff | override)
    return color_map

def plot_mu_phase_diagram(
    df, alpha=0.1, element=None, color_override: dict[str, str] = {},
    poly_method: Literal["concave", "segments"] = 'concave',
):
    df = df.query("stable")

    color_map = get_phase_colors(df.phase.unique(), color_override)

    if "refined" in df.columns and poly_method == "segments":
        tdf = get_transitions(df)
        polys = tdf.groupby("phase").apply(
            make_poly,
            variables=["mu", "T"],
        )
    else:
        polys = df.groupby("phase").apply(
            make_concave_poly,
            alpha=alpha,
            variables=["mu", "T"],
        )

    ax = plt.gca()
    for i, (phase, p) in enumerate(polys.items()):
        p.zorder = 1/p.get_extents().size.prod()
        p.set_color(color_map[phase])
        p.set_edgecolor("k")
        # p.set_alpha(.8)
        p.set_label(polys.index[i])
        ax.add_patch(p)

    mus = df["mu"].unique()
    mus = mus[np.isfinite(mus)]
    plt.xlim(mus.min(), mus.max())
    plt.ylim(df["T"].min(), df["T"].max())
    plt.legend(ncols=2)
    plt.xlabel(r"$\Delta\mu$ [eV]")
    plt.ylabel("$T$ [K]")

def plot_1d_mu_phase_diagram(df):
    """Plot a one dimensional isothermal phase diagram."""
    if len(df['T'].unique()) > 1:
        raise ValueError("data contains more than one temperature!")
    sns.lineplot(
        data=df,
        x='mu', y='phi',
        hue='phase',
        style='stable', style_order=[True, False],
    )
    if 'border' not in df.columns: return

    dfa = np.ptp(df['phi'].dropna())
    dfm = np.ptp(df['mu'].dropna())

    for mt, dd in df.query("mu.min()<mu<mu.max() and border").groupby("mu"):
        ft = dd['phi'].iloc[0]
        plt.axvline(mt, color='k', linestyle='dotted', alpha=.5)
        plt.scatter(mt, ft, marker='o', c='k', zorder=10)

        plt.text(mt - .05 * dfm, ft - dfa * .1, rf"$\Delta\mu = {mt:.03f}\,\mathrm{{eV}}$",
                 rotation='vertical', ha='center', va='top')
    plt.xlabel("Chemical Potential Difference [ev]")
    plt.ylabel("Semi-grandcanonical Potential [eV/atom]")

def plot_1d_T_phase_diagram(df):
    """Plot a one dimensional equipotential phase diagram."""
    if len(df.mu.unique()) > 1:
        raise ValueError("data contains more than one chemical potential!")
    sns.lineplot(
        data=df,
        x='T', y='phi',
        hue='phase',
        style='stable', style_order=[True, False],
    )
    if 'border' not in df.columns: return

    dfa = np.ptp(df['phi'].dropna())
    dft = np.ptp(df['T'].dropna())

    for Tt, dd in df.query("T.min()<T<T.max() and border").groupby("T"):
        ft = dd['phi'].iloc[0]
        plt.axvline(Tt, color='k', linestyle='dotted', alpha=.5)
        plt.scatter(Tt, ft, marker='o', c='k', zorder=10)

        plt.text(Tt + .05 * dft, ft + dfa * .1, rf"$T = {Tt:.0f}\,\mathrm{{K}}$", rotation='vertical', ha='center')
    plt.xlabel("Temperature [K]")
    plt.ylabel("Semi-grandcanonical potential [eV/atom]")



from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from typing import Literal
import warnings

import numpy as np
import scipy.optimize as so
try:
    import polyfit
except ImportError:
    polyfit = None

from scipy.constants import Boltzmann, eV
from sklearn.preprocessing import PolynomialFeatures
from sklearn.pipeline import make_pipeline
from sklearn.linear_model import Ridge, Lasso, LinearRegression

kB = Boltzmann / eV


def G_calphad(T, pl, *p):
    with np.errstate(divide="ignore"):
        g = T * np.log(T) * pl + sum(pi * T**i for i, pi in enumerate(p))
    if isinstance(T, np.ndarray) and T.ndim > 0:
        g[np.isclose(T, 0)] = p[0]
    elif np.isclose(T, 0.0):
        g = p[0]
    return g


Interpolation = Callable[[float], float]
"""Generic Interface for a 1D interpolation."""


class Interpolator(ABC):
    """
    This class acts as a factory for an interplation.

    Call :meth:`.fit()` to obtain a specific interpolation.

    Implementations should be hashable and immutable to allow caching in the
    thermodynamics module.
    """

    @abstractmethod
    def fit(self, x, y) -> Interpolation:
        pass


# subclasses for type hinting only; interface the same
class TemperatureInterpolator(Interpolator):
    pass


class ConcentrationInterpolator(Interpolator):
    pass


@dataclass(frozen=True, eq=True)
class PolyFit(TemperatureInterpolator, ConcentrationInterpolator):
    nparam: int | Literal["auto"]
    """Number of parameters, if "auto" fit a 10 parameter polynomial under L1 and discard parameters <1e-10, then refit."""
    regularizer_strength: float = 1e-8
    """Strength of L2-norm regularization."""
    enforce_curvature: bool = False
    """Ensure that the interpolation has negative curvature as expected for thermodynamic potentials."""

    def fit(self, x, y):
        x = np.asarray(x)
        y = np.asarray(y)
        if self.nparam == "auto":
            reg = make_pipeline(
                    PolynomialFeatures(10),
                    Lasso(self.regularizer_strength, fit_intercept=False)
            )
            reg.fit(x.reshape(-1, 1), y)
            nparam = sum(abs(reg.steps[-1][1].coef_) > 1e-10)
        else:
            nparam = self.nparam
        if not self.enforce_curvature or polyfit is None:
            if self.enforce_curvature:
                warnings.warn("enforce_curvature=True is only supported when the `polyfit` package from PyPI is installed. "
                              "Falling back to regular fitting.")
            reg = make_pipeline(
                    PolynomialFeatures(nparam - 1),
                    Ridge(self.regularizer_strength, fit_intercept=False)
            )
            reg.fit(x.reshape(-1, 1), y)
            coef = reg.steps[-1][1].coef_[::-1]
        else:
            reg = polyfit.PolynomRegressor(
                    nparam, lam=self.regularizer_strength
            ).fit(
                    x.reshape(-1, 1), y,
                    constraints={0: polyfit.Constraints(curvature="concave")}
            )
            coef = reg.coeffs_[::-1]
        return np.poly1d(coef)


@dataclass(frozen=True, eq=True)
class SGTE(TemperatureInterpolator):
    nparam: int

    def __post_init__(self):
        assert self.nparam > 1, "Must fit at least two parameters!"

    def fit(self, x, y):
        parameters, *_ = so.curve_fit(G_calphad, x, y, p0=[0] * self.nparam)
        return lambda x: G_calphad(x, *parameters)


@dataclass(frozen=True, eq=True)
class RedlichKister(ConcentrationInterpolator):
    """
    Fits the "enthalpic" part of a Redlich-Kister expansion, i.e. without the
    ideal configuration entropy.
    """

    nparam: int

    def __post_init__(self):
        assert self.nparam > 0, "Must fit at least one parameter!"

    def fit(self, c, f):
        """
        Beware: You need to manually remove the entropy if included in f.
        """
        # FIXME: assumes terminals are unique
        I = c.argsort()
        f = f[I]
        c = c[I]
        assert np.isclose(c[0], 0) and np.isclose(c[-1], 1), "Must include terminals when fitting Redlich-Kister!"
        f0 = f[0]
        df = f[-1] - f[0]
        f -= f0 + df * c
        nparam = min(self.nparam, len(c) - 2)
        rk_parameters, _ = so.curve_fit(RedlichKisterInterpolation._eval_mix, c, f, p0=np.zeros(nparam))
        return RedlichKisterInterpolation(df, f0, rk_parameters)


@dataclass(frozen=True)
class RedlichKisterInterpolation:
    df: float
    """Change in mixing "enthalpy" across composition range."""
    f0: float
    """Absolute "enthalpy" at concentration 0"""
    rk_parameters: np.ndarray[float]
    """Redlich-Kister parameters."""

    @staticmethod
    def _eval_mix(x, *L):
        pre = x * (1 - x)
        if isinstance(x, np.ndarray):
            vam = np.vander((2 * x - 1), N=len(L), increasing=True)
            return pre * np.einsum("ij,j->i", vam, L)
        else:
            return pre * sum(Li * (2 * x - 1) ** i for i, Li in enumerate(L))

    @staticmethod
    def _eval_mix_derivative(x, *L):
        pre = x * (1 - x)
        xi = 2 * x - 1
        x2 = xi**2
        ds = np.stack([(2 * k * pre - x2) * xi ** (k - 1) for k in range(len(L))])
        if len(ds.shape) == 1:
            return (L * ds).sum()
        else:
            return np.transpose(L) @ ds

    # def fit_derivative(self, c, mu, f0=0, c0=0):
    #     """
    #     Beware: You need to manually remove the entropy if included in mu.
    #     """
    #     # optimization works better if all parameters are on one scale
    #     # df tends to be eV, but *L 1-10meV
    #     # so just center on a rough guess and fit difference to it for df
    #     df_guess = np.median(mu)
    #     nparam = min(len(self.rk_parameters), len(c) - 2)
    #     (self.df, *self.rk_parameters), *_ = so.curve_fit(
    #             lambda c, df, *L: df_guess + df + self._eval_mix_derivative(c, *L),
    #             c, mu, p0=np.zeros(nparam + 1)
    #     )
    #     self.df += df_guess
    #     self.f0 = f0 - self(c0)
    #     return self

    def __call__(self, c):
        return self._eval_mix(c, *self.rk_parameters) + self.f0 + self.df * c


@dataclass(frozen=True, eq=True)
class StitchedFit(TemperatureInterpolator):
    """
    An interpolator with more control over the extrapolation regions.
    """

    interpolating: TemperatureInterpolator = SGTE(4)
    # use the interpolating fit for lower temps too, i.e. extrapolate
    low: TemperatureInterpolator | None = None
    # use a straight line (constant entropy) for higher temperatures
    upp: TemperatureInterpolator | None = PolyFit(2)

    """How many samples near the edges to use to fit the extrapolating interpolator."""
    edge: int = 10

    def fit(self, t, f):
        tmin = t.min()
        tmax = t.max()
        mid = self.interpolating.fit(t, f)
        low = None
        upp = None
        if self.low is not None:
            low = self.low.fit(t[: self.edge], f[: self.edge])
        if self.upp is not None:
            upp = self.upp.fit(t[-self.edge :], f[-self.edge :])

        def interpolation(t):
            t = np.array(t)
            f = mid(t)
            if low is not None:
                f = np.where(t < tmin, low(t), f)
            if upp is not None:
                f = np.where(t > tmax, upp(t), f)
            if f.ndim == 0:
                f = f.item()
            return f

        return interpolation



"""
Error propagation for pedestrians.
"""

from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass, field

from .phases import Phase


@dataclass(frozen=True)
class RandomlyShiftedPhase(Phase):
    """
    Add a constant noise to an arbitrary phase.  Purely for error estimation via bootstrapping.
    """

    name: str
    phase: Phase
    noise: float = 1e-3
    shift: float = field(init=False)

    def __post_init__(self, *args, **kwargs):
        object.__setattr__(self, "shift", np.random.normal(loc=0, scale=self.noise))

    def semigrand_potential(self, T, mu):
        return self.phase.semigrand_potential(T, mu) + self.shift

    def concentration(self, T, mu):
        return self.phase.concentration(T, mu)


def _resample_borders_once(phases, Ts, dmus, noise_norm, run=0):
    noise_phases = [RandomlyShiftedPhase(p.name, p, noise=noise_norm) for p in phases]
    df = calc_phase_diagram(noise_phases, Ts, dmus, refine=True).query("border and -inf<mu<inf")
    df["run"] = run
    return df


def resample_borders(phases, Ts, dmus, noise_norm=1e-3, n=100, cores=20):
    with ProcessPoolExecutor(max_workers=cores) as pool:
        return pd.concat(
            list(pool.map(_resample_borders_once, [phases] * n, [Ts] * n, [dmus] * n, [noise_norm] * n, range(n))),
            ignore_index=True,
        )



"""
Calculates phase diagrams from sets of Phases.
"""

from functools import partial
import numbers
import warnings

import numpy as np
import pandas as pd
import scipy.optimize as so
from scipy.spatial import Delaunay
from typing import Iterable

from scipy.constants import Boltzmann, eV

kB = Boltzmann / eV

from .phases import Phase, AbstractLinePhase


def find_one_point(phase1, phase2, potential, var_range):
    """
    Find a exact phase transition between to phases.

    Args:
        phase1, phase2 (:class:`landau.phase.Phase`):
            the two phases
        potential (callable):
            function that given a phase and an intensive state variable returns a thermodynamic potential
        var_range (tuple of float):
            interval to search for the transition
    """
    return so.root_scalar(
        lambda x: potential(phase1, x) - potential(phase2, x), bracket=var_range, x0=np.mean(var_range), xtol=1e-6
    ).root


def find_mu_one_point(phase1, phase2, mu_range, T):
    """
    Extract chemical potential of a single phase equilibrium.
    """
    mu = find_one_point(phase1, phase2, lambda p, mu: p.semigrand_potential(T, mu), mu_range)
    return [
        {"mu": mu, "phi": phase1.semigrand_potential(T, mu), "c": phase1.concentration(T, mu), "phase": phase1.name},
        {"mu": mu, "phi": phase2.semigrand_potential(T, mu), "c": phase2.concentration(T, mu), "phase": phase2.name},
    ]


def find_T_one_point(phase1, phase2, T_range, mu):
    T = find_one_point(phase1, phase2, lambda p, T: p.semigrand_potential(T, mu), T_range)
    return [
        {"T": T, "phi": phase1.semigrand_potential(T, mu), "c": phase1.concentration(T, mu), "phase": phase1.name},
        {"T": T, "phi": phase2.semigrand_potential(T, mu), "c": phase2.concentration(T, mu), "phase": phase2.name},
    ]


def find_all_points(stable_df, phases, by="mu"):
    """
    Map find_one_point over all estimated equilibria at a given T or mu.
    """
    assert by in ["T", "mu"], "Wrong by value"
    stable_df = stable_df.sort_values(by).reset_index(drop=True)
    boundary_guesses = stable_df.index[(stable_df.phase != stable_df.phase.shift(-1).ffill())]
    if by == "mu":
        boundaries = [
            # {"mu": -np.inf, "c": 0, "phase": stable_df.phase.iloc[0]},
            # {"mu": np.inf, "c": 1, "phase": stable_df.phase.iloc[-1]},
        ]
    else:
        boundaries = [
            # {"T": stable_df["T"].min(), "c": stable_df.c.iloc[0], "phase": stable_df.phase.iloc[0]},
            # {"T": stable_df["T"].max(), "c": stable_df.c.iloc[-1], "phase": stable_df.phase.iloc[-1]},
        ]
    for g in boundary_guesses:
        r1 = stable_df.loc[g]
        r2 = stable_df.loc[g + 1]
        p1 = phases[r1.phase]
        p2 = phases[r2.phase]
        match by:
            case "mu":
                mus = sorted([r1.mu, r2.mu])
                refinements = find_mu_one_point(
                    p1,
                    p2,
                    mus,
                    r1["T"],
                )
            case "T":
                Ts = sorted([r1["T"], r2["T"]])
                refinements = find_T_one_point(
                    p1,
                    p2,
                    Ts,
                    r1["mu"],
                )
        # the find*point functions find the point where the potentials of the two phases are equal, but a third phase
        # could be lower in potential, so only add the refined points if they are truly stable
        # FIXME: technically this only needs to be done once, since the refinements share T/mu
        for phase in phases.values():
            if phase.name in (r1.phase, r2.phase):
                continue
            for point in refinements:
                T = point.get("T", r1["T"])
                mu = point.get("mu", r1["mu"])
                if phase.semigrand_potential(T, mu) < point["phi"]:
                    break
            else:
                continue
            break
        else:
            boundaries.extend(refinements)
    df = pd.DataFrame(boundaries)
    if len(df) > 0:
        df = df.sort_values(by)
    return df


def find_triangle(phases, cand):
    """
    Assumes two of the three points in cand are of the same phase.

    Args:
        phases (Phases): all phases
        cand (dataframe): subset of phase diagram dataframe of the three points of the triangle
    """
    # one of p1, p2 will be the peak of the triangle, the other one the center of the base
    p1, p2 = cand.groupby("phase")[["T", "mu"]].mean().to_numpy()

    def project(t):
        T, mu = p1 + (p2 - p1) * t
        return T, mu

    phase1, phase2 = [phases[p] for p in cand.phase.unique()]
    try:
        t = find_one_point(phase1, phase2, lambda phase, t: phase.semigrand_potential(*project(t)), (0, 1))
    except ValueError:
        warnings.warn(f"Failed to refine triangle between {p1} and {p2} of phases {cand.phase.unique()}!", stacklevel=2)
        return []
    T, mu = p1 + (p2 - p1) * t
    if T < 0:
        return []
    phi = phase1.semigrand_potential(T, mu)
    # check that no other phase is lower than the refined boundary
    if any(p.semigrand_potential(T, mu) < phi for p in phases.values() if p.name not in (phase1.name, phase2.name)):
        # TODO: could try and refine here on the boundary (p1, more_stable_phase) and (p2, more_stable_phase)
        return []
    return [
        {"T": T, "mu": mu, "phi": phases[p].semigrand_potential(T, mu), "c": phases[p].concentration(T, mu), "phase": p}
        for p in cand.phase.unique()
    ]


def refine_phase_diagram(df, phases, min_c=0, max_c=1):
    """Add additional points to a coarse phase diagram by searching for exact transitions."""
    udf = df.query("not stable").reset_index(drop=True)
    udf["border"] = False
    df = df.query("stable").reset_index(drop=True)
    df["border"] = False
    df["refined"] = "no"
    data = [df, udf]
    multiple_mus = len(df["mu"].unique()) > 1
    multiple_ts = len(df["T"].unique()) > 1
    if multiple_mus and multiple_ts:
        # declare edges of the sampling window as borders so to not confuse get_transitions, debatably hacky
        df.loc[df["T"] == df["T"].min(), "border"] = True
        df.loc[df["T"] == df["T"].max(), "border"] = True
        # left and right edges as well, set here to +-inf to make sure the
        # cluster algo below separates top and left and right edges, even
        # hackier
        left = df.loc[df["mu"] == df["mu"].min()][["phase", "T"]]
        left["mu"] = -np.inf
        left["c"] = min_c
        left["border"] = True
        left["stable"] = True
        data.append(left)
        right = df.loc[df["mu"] == df["mu"].max()][["phase", "T"]]
        right["mu"] = +np.inf
        right["c"] = max_c
        right["border"] = True
        right["stable"] = True
        data.append(right)
        # Main idea:
        # - tessellate input points
        # - count the number of unique phases in each triangle
        # - if > 1 there must be at least one phase transition in the triangle
        # - find_triangle assumes (erroneously) that it can be found on the vector connecting the peak of the triangle
        # to the center of the base
        # - if = 3 there's probably a triple point in there (but doesn't need to be actually, should check that)
        dela = Delaunay(df[["mu", "T"]])
        coex = df.phase.to_numpy()[dela.simplices]
        phase_counts = np.array([len(set(x)) for x in coex])
        line_candidates = [df.iloc[i] for i in dela.simplices[phase_counts == 2]]
        trip_candidates = [df.iloc[i] for i in dela.simplices[phase_counts == 3]]
        # you'd think this to be faster, but somehow un/pickling the phases is very slow, likely because it has to do it
        # for each triangle and involves fitting the phases all over
        # with ProcessPoolExecutor(4) as pool:
        #     ddf = pool.map(partial(find_triangle, phases), line_candidates)
        ddf = map(partial(find_triangle, phases), line_candidates)
        ddf = pd.DataFrame(sum(ddf, []))
        ddf["stable"] = True
        ddf["border"] = True
        ddf["refined"] = "delaunay"
        data.append(ddf)

        def refine_triples(tr):
            T, mu = tr[["T", "mu"]].mean()
            p1, p2, p3 = (phases[p] for p in tr.phase.unique())

            def triplemin(x):
                T, mu = x
                phi1 = p1.semigrand_potential(T, mu)
                phi2 = p2.semigrand_potential(T, mu)
                phi3 = p3.semigrand_potential(T, mu)
                return abs(phi1 - phi2) + abs(phi2 - phi3) + abs(phi3 - phi1)

            T, mu = so.fmin(triplemin, (T, mu), disp=False)
            if T < 0:
                return []
            return [
                {
                    "T": T,
                    "mu": mu,
                    "phi": phases[p].semigrand_potential(T, mu),
                    "c": phases[p].concentration(T, mu),
                    "phase": p,
                }
                for p in tr.phase.unique()
            ]

        tdf = []
        for tri in trip_candidates:
            tdf.extend(refine_triples(tri))
        tdf = pd.DataFrame(tdf)
        tdf["stable"] = True
        tdf["border"] = True
        tdf["refined"] = "delaunay-triple"
        data.append(tdf)
    else:
        if multiple_mus:
            mdf = (
                df.groupby("T", group_keys=True)
                .apply(find_all_points, phases=phases, by="mu", include_groups=True)
                .reset_index()
            )
            mdf["stable"] = True
            mdf["border"] = True
            mdf["refined"] = "mu"
            data.append(mdf.drop("level_1", axis="columns"))
        if multiple_ts:
            Tdf = (
                df.groupby("mu", group_keys=True)
                .apply(find_all_points, phases=phases, by="T", include_groups=True)
                .reset_index()
            )
            Tdf["stable"] = True
            Tdf["border"] = True
            Tdf["refined"] = "T"
            data.append(Tdf.drop("level_1", axis="columns"))
    return pd.concat(data, ignore_index=True)


def guess_mu_range(phases: Iterable[Phase], T: float, samples: int, tolerance: float = 1e-2):
    """Guess chemical potential window from the ideal solution.

    Searches numerically for chemical potentials which stabilize
    concentrations close to 0 and 1 and then use the concentrations
    encountered along the way to numerically invert the c(mu) mapping.
    Using an even c grid with mu(c) then yields a decent sampling of mu
    space so that the final phase diagram is described everywhere equally.

    Args:
        phases: list of phases to consider
        T: temperature at which to estimate mu(c)
        samples: how many mu samples to return

    Returns:
        array of chemical potentials that likely cover the whole concentration space
    """
    # TODO: this can be used immediately also for the actual phase diagram
    # calculation: keep track of which phase is the most likely
    import scipy.optimize as so
    import scipy.interpolate as si
    import numpy as np
    # semigrand canonical "average" concentration
    # use this to avoid discontinuities and be phase agnostic

    def c(mu):
        phis = np.array([p.semigrand_potential(T, mu) for p in phases])
        conc = np.array([p.concentration(T, mu) for p in phases])
        phis -= phis.min(axis=0)
        beta = 1 / (kB * T)
        prob = np.exp(-beta * phis)
        prob /= prob.sum(axis=0)
        ci = (prob * conc).sum(axis=0)
        return ci

    resi = so.minimize(lambda x: +c(x[0]), x0=[0], tol=tolerance, method="BFGS")
    resa = so.minimize(lambda x: -c(x[0]), x0=[0], tol=tolerance, method="BFGS")
    mu0 = resi.x[0]
    mu1 = resa.x[0]
    if mu0 == mu1:
        if tolerance > 1e-7:
            return guess_mu_range(phases, T, samples, tolerance/10)
        raise ValueError(
                "chemical potential range degenerate! Check that phases that not all phases have the same fixed "
                "concentration!"
        )
    mm = np.linspace(mu0, mu1, samples)
    cc = c(mm)
    c0 = min(cc) + tolerance
    c1 = max(cc) - tolerance
    return si.interp1d(cc, mm)(np.linspace(c0, c1, samples)), c0, c1


def calc_phase_diagram(
    phases: Iterable[Phase],
    Ts: Iterable[float] | float,
    mu: Iterable[float] | float | int,
    refine: bool = True,
    keep_unstable: bool = False,
):
    """
    Calculate phase diagram at given sampling points.

    Args:
        phases (iterable of Phases)
        Ts (iterable of floats): sampling points in temperature
        mu (iterable of floats): sampling points in chemical potential; if int
            guess sampling points with guess_mu_range at max(Ts)
        refine (bool): add additional sampling points at exact phase transitions
        keep_unstable (bool): only keep entries of stable phases, otherwise keep entries of all phases at all sampling points

    Returns:
        dataframe of phase points
    """
    if not isinstance(Ts, Iterable):
        Ts = [Ts]
    phases = {p.name: p for p in phases}
    if isinstance(mu, numbers.Integral) and mu != 0:
        # we would often pass mu=0 to calculate a fixed mu, temperature only diagram and it'd be a bit annoying to pass
        # mu=0.0 all the time, so we special case as above
        try:
            mu, min_c, max_c = guess_mu_range(phases.values(), max(Ts), int(mu))
        except ValueError:
            if all(isinstance(p, AbstractLinePhase) for p in phases.values()):
                raise ValueError(
                        "Cannot guess chemical potential range of line phases with all the same concentration!"
                ) from None
            raise
    elif refine:
        min_c, max_c = None, None

    def get(s, T):
        phi = s.semigrand_potential(T, mu)
        return {"T": T, "phase": s.name, "phi": phi, "mu": mu, "c": s.concentration(T, mu)}

    pdf = pd.DataFrame([get(s, T) for s in phases.values() for T in Ts])
    pdf = pdf.explode(["mu", "phi", "c"]).infer_objects().reset_index(drop=True)
    pdf["stable"] = False
    pdf.loc[pdf.groupby(["T", "mu"], group_keys=False).phi.idxmin(), "stable"] = True
    if refine:
        min_c = pdf.c.min()
        max_c = pdf.c.max()
        pdf = refine_phase_diagram(pdf, phases, min_c=min_c, max_c=max_c)
    pdf["f"] = pdf.phi + pdf.mu * pdf.c

    def sub(dd):
        dd = dd.query("-inf<mu<inf")
        c0 = dd.c.min()
        c1 = dd.c.max()
        f0 = dd.query("c==@c0").f.min()
        f1 = dd.query("c==@c1").f.min()
        return dd.f - (f0 * (1 - dd.c) + f1 * dd.c)

    fex = pdf.groupby("T", group_keys=False).apply(sub, include_groups=False)
    if len(Ts) > 1:
        pdf["f_excess"] = fex
    else:
        # thank you pandas, this saved me -10min of my life.
        pdf["f_excess"] = fex.T
    if not keep_unstable:
        pdf = pdf.query("stable")
    return pdf


from sklearn.cluster import AgglomerativeClustering


def reduce(dd):
    dd = dd.sort_values("c")
    return pd.Series(
        {
            "transition": "-".join(dd.phase.tolist()),
            "c": dd.c.tolist(),
            "phase": dd.phase.tolist(),
        }
    )


def cluster(dd, eps=0.01):
    t = dd["T"]
    # Guard against isothermal segments
    if t.min() != t.max():
        t = (t - t.min()) / (t.max() - t.min())
    ids = pd.Series(np.zeros_like(dd.index), index=dd.index)
    # picking eps is a pain and HDBSCAN not available in my env yet
    # cluster = DBSCAN(eps=eps, min_samples=3)
    cluster = AgglomerativeClustering(
        n_clusters=None,
        # FIXME: hand optimized value; smaller values tend to partition the
        # same transition too often
        distance_threshold=0.5,
        linkage="single",
    )
    # on the left and right side of the phase diagram refining adds points with
    # mu +- inf, which chokes the cluster methods, but we know they should be
    # their own segments, so special case them below
    F = np.isfinite(dd.mu)
    if F.any() and sum(F) >= 2:
        ids.loc[F] = cluster.fit_predict(np.transpose([t.loc[F], dd.c.loc[F], dd.mu.loc[F]]))
    m = ids.max()
    ids.loc[dd.mu == +np.inf] = m + 1
    ids.loc[dd.mu == -np.inf] = m + 2
    return ids


def get_transitions(df):
    """
    Identify "continuous" two-phase transition lines in mu/T space, i.e. transitions between the same two phases and along which mu/T are continuous.

    Useful for plotting below, but potentially also to augment the existing refining routines and
    acquire additional Free energies from calphy/etc. to improve the diagram.
    """
    bdf = df.query("border")
    # go from a table of mu/c/T points that are on the phase boundaries to a table where the two points that are at the same mu/T are grouped together
    # use this information to add 'transition' column; handles also the case where border points are at mu=+-inf, there we have only one point
    tdf = bdf.groupby(["mu", "T"])[["c", "phase"]].apply(reduce)
    # immediately explode again to go back to our familiar representation, but now with the added 'transition' column
    tdf = tdf.reset_index().explode(["c", "phase"]).infer_objects().reset_index(drop=True)

    # cluster points that are assigned as one transition, because the same transition can appear multiple times in "disconnected" manner in a phase
    # diagram, e.g. a solid solution in contact with the melt interrupted by a higher melting intermetallic
    tdf["transition_unit"] = tdf.groupby("transition", group_keys=False).apply(cluster, include_groups=False)
    tdf["border_segment"] = tdf[["transition", "transition_unit"]].apply(
        lambda r: "_".join(map(str, r.tolist())), axis="columns"
    )

    return tdf



from .phases import (
    LinePhase,
    TemperatureDepandantLinePhase,
    TemperatureDependentLinePhase,
    IdealSolution,
    RegularSolution,
    InterpolatingPhase,
)

from .plot import plot_phase_diagram

try:
    from ._version import __version__
except ImportError:
    __version__ = "dev"
